# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['core', 'core.cli', 'core.executor', 'core.lib', 'core.test']

package_data = \
{'': ['*']}

install_requires = \
['deprecated>=1.2.13,<2.0.0',
 'dill>=0.3.5,<0.3.6',
 'kraken-common>=0.5.2,<0.6.0',
 'networkx>=2.6.3,<2.7.0',
 'nr-io-graphviz>=0.1.1,<0.2.0',
 'nr-stream>=1.1.0,<2.0.0',
 'setuptools>=33.1.0',
 'termcolor>=1.1.0,<2.0.0',
 'typeapi>=0.2.2,<0.3.0',
 'typing-extensions>=3.10.0.2']

extras_require = \
{'testing': ['pytest>=6.0.0']}

entry_points = \
{'console_scripts': ['kraken = kraken.core.cli.main:main'],
 'pytest11': ['kraken.core.test = kraken.core.test']}

setup_kwargs = {
    'name': 'kraken-core',
    'version': '0.11.2',
    'description': '',
    'long_description': '<img align="center" src="docs/content/img/title.png">\n\n# kraken-core\n\n[![Python application](https://github.com/kraken-build/kraken-core/actions/workflows/python-package.yml/badge.svg)](https://github.com/kraken-build/kraken-core/actions/workflows/python-package.yml)\n[![PyPI version](https://badge.fury.io/py/kraken-core.svg)](https://badge.fury.io/py/kraken-core)\n\nKraken is a task orchestration system with a focus on ease of configuration and flexibility. It\'s main objectives are\nto reduce boilerplate and configuration overhead in software projects and to model and abstract complex workflows in\nthe software development lifecycle to allow developers to focus on what really matters.\n\nRead more about the Kraken core package in the [documentation][].\n\n  [documentation]: https://kraken-build.github.io/kraken-core/\n',
    'author': 'Niklas Rosenstein',
    'author_email': 'rosensteinniklas@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
